"""
http adapter
"""
import requests
import requests_cache

from flume import logger, moment
from flume.adapters.adapter import adapter
from flume.exceptions import FlumineException
from flume.point import Point


class http(adapter):
    """
    http adapter
    """

    name = 'http'

    def __init__(self,
                 url=None,
                 method='GET',
                 headers=None,
                 time='time',
                 filter=None,
                 follow_link=True,
                 cache=None):
        self.url = url
        self.method = method
        self.headers = headers
        self.time = time
        self.filter = filter
        self.follow_link = follow_link
        self.cache = cache

    def read(self):
        """
        read points by issuing an HTTP request and pushing the response as
        individual points into the flume pipeline
        """
        url = self.url

        while True:
            if self.cache is not None:
                requests_cache.install_cache(self.cache)

            response = requests.request(self.method, url, headers=self.headers)

            if response.status_code != 200:
                raise FlumineException(response.text)

            data = response.json()

            points = []
            for point in data:
                # convert to a flume Point
                point = Point(**point)

                if self.time in point:
                    point.time = moment.date(point[self.time])
                    points.append(point)

                else:
                    logger.warn('point missing time field "%s"' % self.time)

            yield points

            if not self.follow_link:
                break

            if 'link' in response.headers and 'next' in response.links:
                url = response.links['next']['url']

            else:
                break

    def write(self, points):
        payload = []
        for point in points:
            payload.append(point.json())

        if self.cache is not None:
            requests_cache.install_cache(self.cache)

        response = requests.request(self.method,
                                    self.url,
                                    headers=self.headers,
                                    json=payload)

        if response.status_code != 200:
            raise FlumineException('received bad response with %d: %s' %
                                   (response.status_code, response.text))

    def eof(self):
        pass



"""
tail processor
"""
from flume import node


class tail(node):
    """
    # tail

    The tail processor is used to only keep the last N elements and throw
    away all other remaining points

    usage: ... | tail(N) | ...

    """

    name = 'tail'

    def __init__(self, howmany):
        node.__init__(self)
        self.howmany = howmany

    def loop(self):
        points = []

        while self.running:
            points += self.pull()
            points = points[-self.howmany:]

        self.push(points[-self.howmany:])



"""
put processor
"""
from flume import node, reducer


class put(node):
    """
    # put

    The put processor is used to to add fields to any point that passes through
    this node in the pipeline. The fields added can be static or the result of
    applying a **reducer**.

    usage: ... | put(field=value, ..., field=count()) | ...

    """

    name = 'put'

    def __init__(self, **fields):
        node.__init__(self, **fields)
        self.fields = fields

    def loop(self):
        while self.running:
            points = self.pull()

            for point in points:
                # call update every time through here
                for (key, value) in self.fields.items():
                    if isinstance(value, reducer):
                        value.update(point)
                        point[key] = value.result()
                    else:
                        point[key] = value

            self.push(points)



"""
stdio adapter
"""
import sys

from flume.adapters.adapter import adapter
from flume.adapters import streamers
from flume.point import Point
from flume import logger, moment


class stdio(adapter):
    """
    stdio adapter

    Used to read/write points from standard IO which includes the stdin,
    stdout and stderr, but can also be used to read/write points to a
    file on the filesystem
    """

    name = 'stdio'

    stdout = sys.stdout
    stdin = sys.stdin

    def __init__(self,
                 time='time',
                 format='jsonl',
                 file=None,
                 **kwargs):
        self.time = time
        self.streamer = streamers.get_streamer(format, **kwargs)
        self.file = file

    def __read(self, stream):
        for point in self.streamer.read(stream):
            if self.time in point.keys():
                point['time'] = moment.date(point[self.time])

                if self.time != 'time':
                    del point[self.time]

                yield [Point(**point)]

            else:
                yield [Point(**point)]

    def read(self):
        if self.file is None:
            for point in self.__read(stdio.stdin):
                yield point

        else:
            with open(self.file, 'r') as stream:
                for point in self.__read(stream):
                    yield point

    def write(self, points):
        if self.file is None:
            self.streamer.write(stdio.stdout, points)

        else:
            with open(self.file, 'a') as stream:
                self.streamer.write(stream, points)

    def eof(self):
        self.streamer.eof(stdio.stdout)



"""
common module

- defines various basic elements of what make up a flume pipeline
- handles all of the fancy logic for the dot(.) and pipe(|) notation when
  construction pipelines
"""
import bisect
import logging
import sys
import threading

from Queue import Empty
from Queue import Queue as queue
from flume import logger
from flume import moment

from flume.exceptions import FlumineException
from flume.point import Point

__SOURCES = {}
__PROCS = {}
__SINKS = {}


def build_node_func(which_node):
    """
    internally used method to attach new nodes to existing ones in the flume
    pipeline when registering such elements
    """

    def node_func(self, *args, **kwargs):
        """
        internal method responsible for hooking up the nodes in a flume
        pipeline
        """

        node_instance = which_node(*args, **kwargs)

        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[queue()])

        source = self.source if hasattr(self, 'source') else self
        node.init_node(node_instance,
                       inputs=self.outputs,
                       outputs=[queue()],
                       parent=self,
                       source=source)

        return node_instance

    return node_func

def register_source(source_class):
    """
    registers the source within flume and makes sure to setup any existing
    processors and sinks as chainable methods
    """
    if source_class.name in __SOURCES.keys():
        raise FlumineException('"%s" source already registered' % source_class.name)

    logger.debug('register_source %s', source_class.name)
    __SOURCES[source_class.name] = source_class

    for name in __PROCS.keys():
        setattr(source_class, name, build_node_func(__PROCS[name]))

    for sink_name in __SINKS.keys():
        setattr(source_class, sink_name, build_node_func(__SINKS[sink_name]))


def register_sink(sink_class):
    """
    registers the sink with flume and makes sure to hook itself up on any
    existing procs or sources as a chainable method
    """
    if sink_class.name in __SINKS.keys():
        raise FlumineException('"%s" sink already registered' % sink_class.name)

    logger.debug('register_sink %s', sink_class.name)
    __SINKS[sink_class.name] = sink_class

    for source in __SOURCES.values():
        setattr(source, sink_class.name, build_node_func(sink_class))

    for proc in __PROCS.values():
        setattr(proc, sink_class.name, build_node_func(sink_class))


def register_proc(proc_class):
    """
    registers the proc_class with flume and makes sure to hook any existing
    procs and sinks as chainable methods
    """
    if proc_class.name in __PROCS.keys():
        raise FlumineException('"%s" proc already registered' % proc_class.name)

    logger.debug('register_proc %s', proc_class.name)
    __PROCS[proc_class.name] = proc_class

    for source_class in __SOURCES.values():
        setattr(source_class, proc_class.name, build_node_func(proc_class))

    for other_proc_class in __PROCS.values():
        setattr(other_proc_class,
                proc_class.name,
                build_node_func(proc_class))
        setattr(proc_class,
                other_proc_class.name,
                build_node_func(other_proc_class))

    for sink_class in __SINKS.values():
        setattr(proc_class, sink.name, build_node_func(sink_class))

class node(object):
    """
    base flume node class
    """

    name = None

    EOF = Point(__eof=True)

    def __init__(self, *args, **kwargs):
        """
        the base node __init__ is just here to be record the arguments and
        keyword arguments used by each node
        """
        self.args = args
        self.kwargs = kwargs
        self.inited = True
        self.running = True
        self.exc_info = queue()
        self.inputs_index = None

    def init_node(self,
                  source=None,
                  parent=None,
                  inputs=None,
                  outputs=None):
        self.source = source
        self.inputs = inputs
        self.outputs = outputs
        self.parent = parent

    def pull(self, wait=True, timeout=None):
        """
        return a batch of points from the upstream procs and guarantee all
        points are sorted in order otherwise
        """
        times = []
        result = []

        if self.inputs is not None:

            if self.inputs_index is None:
                self.inputs_index = {}
                index = 0
                for input in self.inputs:
                    self.inputs_index[input] = index
                    index += 1

            for input in self.inputs:
                input_index = self.inputs_index[input]

                try:
                    points = input.get(wait)
                except Empty:
                    points = []

                for point in points:
                    point = Point(**point.copy())

                    if point == self.EOF:
                        self.inputs.remove(input)
                        continue

                    point.__meta__.flume_path += ('.%s[%s]' % (self.name, input_index))

                    if 'time' not in point.keys():
                        result.append(point)

                    else:
                        index = bisect.bisect_left(times, point.time)
                        times.insert(index, point.time)
                        result.insert(index, point)

        if self.inputs is None or len(self.inputs) == 0:
            self.running = False

        logger.debug('%s pulling %s points', self, len(result))
        return result

    def push(self, points):
        """
        push the provided points to the outputs of this node so that other
        nodes downstream can receive it
        """
        if not isinstance(points, list):
            points = [points]

        logger.debug('%s pushing %s points', self, len(points))

        for point in points:
            for output in self.outputs:
                if point == self.EOF:
                    output.put([self.EOF])

                else:
                    output.put([Point(**point.copy())])

    def __ror__(self, other):
        """

        """
        return node.__or__(splitter(flumes=other), self)

    def __or__(self, other):
        """
        enable the ability to use the bitwise or operator to emulate the pipe
        notation between flume nodes
        """
        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[queue()])

        source = self.source if hasattr(self, 'source') else self

        if isinstance(other, tuple):
            # split the stream!
            other = splitter(flumes=list(other))

        node.init_node(other,
                       inputs=self.outputs,
                       outputs=[queue()],
                       parent=self,
                       source=source)
        return other

    def loop(self):
        """
        every node should implement a loop method which is where they can handle
        the pushing/pulling of points
        """
        raise FlumineException('Each node should implement the loop method')

    def run(self):
        """
        node run method used by flume internally to manage thread execution
        """
        try:
            self.loop()
            self.exc_info.put(None)

        except:
            self.exc_info.put(sys.exc_info())

        finally:
            # must make sure to push EOF even if there was a failure
            self.push(self.EOF)

    def execute(self,
                wait=True,
                debug=False,
                loglevel=logger.WARN):

        if 'inited' not in self.__dict__ or not self.inited:
            raise FlumineException('node.__init__ was never used')

        # XXX: special case of a single source with no output doing an execute()
        #      on it shouldn't fail
        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[])

        logger.setLogLevel(loglevel)

        # XXX: pooling here ?
        thread = threading.Thread(target=self.run)
        # Daemonize so that when we Ctrl+C the main program then all underlying
        # threads are instantly killed. Currently don't have any concern about
        # cleanly closing resources.
        thread.daemon = True
        thread.start()

        if self.parent:
            self.parent.execute(wait=wait,
                                debug=debug)

        if wait:
            while thread.is_alive():
                # if you don't join with a timeout then you block the parent
                # until the child has completely finished and therefore can't
                # handle any signals in the parent (ie SIGINT)
                thread.join(1)

            exc_info = self.exc_info.get()

            if exc_info is not None:
                raise exc_info[0], exc_info[1], exc_info[2]

    def input_count(self):
        """
        returns how many inputs are currently hooked up to the current node
        """
        if self.parent.name == 'splitter':
            # XXX single point of hackery where we do something different for
            #     the split proc
            if len(self.parent.flumes) == 0:
                # since there's still the split input
                return 1

            else:
                return len(self.parent.flumes)
        else:
            return len(self.inputs)


class sink(node):
    """
    base sink class
    """

    def __init__(self, *args, **kwargs):
        node.__init__(self, *args, **kwargs)


class reducer(object):
    """
    base reducer class
    """

    def update(self, point):
        """
        update method called when there is a new point to be passed to the
        reducer
        """
        raise FlumineException('missing implementation for update method')

    def result(self):
        """
        reducer should return the reduced field value
        """
        raise FlumineException('missing implementation for result method')

class splitter(node):
    """
    # splitter

    The split processor is used internally to handle splitting the flume stream

    usage: ... | (put(...), filter(...), etc) | ...

    """

    name = 'splitter'

    def __init__(self,
                 flumes=None,
                 delay=None):
        node.__init__(self, flumes)
        self.flumes = flumes if flumes is not None else []

        if delay is not None:
            self.delay = moment.duration(delay)
        else:
            self.delay = moment.duration('5s')

        self.flume_outputs = []

    def loop(self):
        # self.input is None when split has no parent
        while self.running:
            points = self.pull()
            self.push(points)

    def execute(self,
                wait=True,
                debug=False,
                loglevel=logging.ERROR):

        def find_root(flume):
            """
            figure out the root node of a flume pipeline
            """
            if hasattr(flume, 'parent') and \
               flume.parent is not None:
                return find_root(flume.parent)
            else:
                return flume

        if not hasattr(self, 'outputs'):
            node.init_node(self, outputs=[])

        if len(self.flumes) != 0:

            forwarder_inputs = self.outputs
            self.outputs = []

            for flume in self.flumes:
                flume_input = queue()
                root = find_root(flume)
                root.parent = None
                root.inputs = [flume_input]

                # don't setup the outputs for a sink as it will never
                # push anything out XXX: could be a bit more elegant
                if not isinstance(flume, sink):
                    output = queue()
                    self.flume_outputs.append(output)
                    flume.outputs = [output]

                self.outputs.append(flume_input)

            from flume.procs import reorder
            forwarder = reorder(delay=self.delay)

            source = self.source if hasattr(self, 'source') else self
            node.init_node(forwarder,
                           inputs=self.flume_outputs,
                           outputs=forwarder_inputs,
                           parent=None,
                           source=source)

            forwarder.execute(wait=False,
                              debug=debug,
                              loglevel=loglevel)

            # start underlying flumes
            for flume in self.flumes:
                flume.execute(wait=False,
                                debug=debug,
                                loglevel=loglevel)

        # override default behavior to execute the underlying flume
        node.execute(self,
                     wait=False,
                     debug=debug,
                     loglevel=loglevel)



import unittest

from robber import expect
from flume import *

from flume.core import register_sink, register_proc, register_source

class CoreTest(unittest.TestCase):
    """
    verifies basic flume flow behaviors
    """

    def test_failure_in_source_propagates_correctly(self):
        try:
            (
                emit(start=-1)
            ).execute()
        except Exception as exception:
            expect(exception.message).to.eq('Unable to parse moment "-1"')

    def test_failure_in_proc_propagates_correctly(self):

        class failme(reducer):

            def __init__(self):
                pass

            def update(self, point):
                raise Exception('failing on purpose')

        try:
            (
                emit(limit=10, start='2010-01-01')
                | put(foo=failme())
            ).execute()
        except Exception as exception:
            expect(exception.message).to.eq('failing on purpose')

    def test_failure_in_sink_propagates_correctly(self):

        try:
            (
                emit(limit=10, start='2010-01-01')
                | write('http', url='bananas')
            ).execute()
        except Exception as exception:
            expect(exception.message).to.contain('Invalid URL \'bananas\'')

    def test_register_existing_source_fails(self):
        try:
            register_source(emit)
        except FlumineException as exception:
            expect(exception).to.contain('"emit" source already registered')

    def test_register_existing_proc_fails(self):
        try:
            register_proc(put)
        except FlumineException as exception:
            expect(exception).to.contain('"put" proc already registered')

    def test_register_existing_sink_fails(self):
        try:
            register_sink(write)
        except FlumineException as exception:
            expect(exception).to.contain('"write" sink already registered')



"""
put proc unittests
"""
import unittest

from robber import expect
from flume import *


class PutTest(unittest.TestCase):

    def test_put_add_a_numerical_field(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo=1)
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1},
            {'foo': 1}
        ])

    def test_put_multiple_fields(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo=1, fizz='buzz')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1, 'fizz': 'buzz'},
            {'foo': 1, 'fizz': 'buzz'}
        ])

    def test_put_multiple_fields_with_multiple_puts(self):
        results = []
        (
            emit(points=[{}, {}])
            | put(foo=1)
            | put(fizz='buzz')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'foo': 1, 'fizz': 'buzz'},
            {'foo': 1, 'fizz': 'buzz'}
        ])



"""
http adapter unittests
"""

import threading
import time
import unittest

import json
import logging
import requests

from flask import Flask, request, Response
from robber import expect

from flume import *

# silence flask logging
logging.getLogger('werkzeug').setLevel(logging.ERROR)

_HOST = '0.0.0.0'
_PORT = 22222
_BASEURL = 'http://%s:%d' % (_HOST, _PORT)


class HTTPTestServer(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)

    def run(self):

        class MyFlask(Flask):

            def __init__(self, *args, **kwargs):
                super(MyFlask, self).__init__(*args, **kwargs)
                self.db = {}

        app = MyFlask(__name__)

        @app.route('/empty')
        def empty():
            return Response('[]',
                            status=200,
                            mimetype='application/json')

        @app.route('/garbage')
        def garbage():
            return Response('garbage',
                            status=200,
                            mimetype='application/json')

        @app.route('/points')
        def points():
            page = int(request.args.get('page', '1'))
            count = int(request.args.get('count', '3'))
            start = request.args.get('start', 'now')
            every = request.args.get('every', '1s')

            start = moment.date(start)
            every = moment.duration(every)
            current_time = start

            result = []
            headers = {}

            if page > 1:
                link = '<%s/points?page=%d&count=%d>; rel="next"' % \
                       (_BASEURL, page - 1, count)
                headers['link'] = link

            for index in range(0, count):
                result.append({
                    'time': moment.datetime_to_iso8601(current_time)
                })
                current_time += every

            return Response(json.dumps(result),
                            status=200,
                            mimetype='application/json',
                            headers=headers)

        @app.route('/store', methods=['PUT', 'GET'])
        def store():
            key = request.args.get('key')

            if request.method == 'PUT':
                app.db[key] = json.loads(request.data)
                return Response(status=200)

            if request.method == 'GET':
                return Response(json.dumps(app.db[key]),
                                status=200,
                                mimetype='application/json')

        @app.route('/shutdown', methods=['POST'])
        def shutdown():
            shutit = request.environ.get('werkzeug.server.shutdown')

            if shutit is None:
                raise RuntimeError('not running with the werkzeug server')

            shutit()

            return Response(status=200)

        try:
            app.run(host=_HOST, port=_PORT)
        except:
            # ignore failures at shutdown
            pass


class HttpTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.server = HTTPTestServer()
        cls.server.start()

        # wait for the server to come up
        # XXX: should set a timeout
        response = None
        while response is None or response.status_code != 200:
            time.sleep(0.1)
            try:
                response = requests.get(_BASEURL + '/empty')
            except:
                pass

    @classmethod
    def tearDownClass(cls):
        requests.post(_BASEURL + '/shutdown')

    def test_http_read_with_failure(self):
        results = []
        try:
            (
                read('http',
                     url=_BASEURL + '/garbage')
                | memory(results)
            ).execute()

            raise Exception('previous code should have failed')
        except ValueError as exception:
            expect(exception.message).to.eq('No JSON object could be decoded')
            expect(results).to.eq([])

    def test_empty_http_read(self):
        results = []
        (
            read('http',
                 url=_BASEURL + '/empty')
            | memory(results)
        ).execute()
        expect(results).to.eq([])

    def test_http_read_a_few_historical_points(self):
        results = []
        (
            read('http',
                 url=_BASEURL + '/points?count=3&start=2010-01-01')
            | memory(results)
        ).execute()

        expect(results).to.eq([
            {'time': '2010-01-01T00:00:00.000Z'},
            {'time': '2010-01-01T00:00:01.000Z'},
            {'time': '2010-01-01T00:00:02.000Z'}
        ])

    def test_http_read_a_few_live_points(self):
        results = []
        (
            read('http',
                 url=_BASEURL + '/points?count=5')
            | memory(results)
        ).execute()

        expect(results).to.have.length(5)

    def test_http_write_a_point_and_read_it_back(self):
        results = []
        (
            emit(limit=1, start='1970-01-01')
            | put(foo='bar')
            | write('http',
                    method='PUT',
                    url=_BASEURL + '/store?key=test1')
        ).execute()

        (
            read('http',
                 method='GET',
                 url=_BASEURL + '/store?key=test1')
            | keep('foo')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'foo': 'bar'}])

    def test_http_write_a_single_payload_and_read_them_back(self):
        results = []
        (
            emit(limit=3, start='1970-01-01')
            | put(count=count())
            | write('http',
                    method='PUT',
                    url=_BASEURL + '/store?key=test2',
                    batch=3)
        ).execute()

        (
            read('http',
                 method='GET',
                 url=_BASEURL + '/store?key=test2')
            | keep('count')
            | memory(results)
        ).execute()

        # if we wrote out multiple paylaods then the end result would be that
        # we would only be able to retrieve the last point
        expect(results).to.eq([
            {'count': 1},
            {'count': 2},
            {'count': 3}
        ])

    def test_http_write_multiple_payloads_and_read_them_back(self):
        results = []
        (
            emit(limit=3, start='1970-01-01')
            | put(count=count())
            | write('http',
                    method='PUT',
                    url=_BASEURL + '/store?key=test2',
                    batch=1)
        ).execute()

        (
            read('http',
                 method='GET',
                 url=_BASEURL + '/store?key=test2')
            | keep('count')
            | memory(results)
        ).execute()

        # because the payload is of a single point we only end up storing the
        # last point that was pushed
        expect(results).to.eq([
            {'count': 3}
        ])

    def test_http_read_can_follow_link_header(self):
        results = []
        (
            read('http',
                 method='GET',
                 url=_BASEURL + '/points?count=2&page=3')
            | reduce(count=count())
            | keep ('count')
            | memory(results)
        ).execute()

        expect(results).to.eq([{'count': 6}])



"""
http adapter unittests
"""

import json
import sys
import unittest

from StringIO import StringIO
from robber import expect

from flume import *


class StdioTest(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        # make sure to set the stdin/stdout hooks in stdio back to the
        # right values
        stdio.stdin = sys.stdin
        stdio.stdout = sys.stdout

    def test_stdio_can_read_from_empty(self):
        stdio.stdin = StringIO('')
        results = []

        (
            read('stdio')
            | memory(results)
        ).execute()
        expect(results).to.eq([])

    def test_stdio_can_read_a_single_point_(self):
        stdio.stdin = StringIO('{"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}')
        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": "bar"}
        ])

    def test_stdio_can_read_multiple_points(self):
        stdio.stdin = StringIO('{"time": "2016-01-01T00:00:00.000Z", "foo": 1}\n' +
                               '{"time": "2016-01-01T00:01:00.000Z", "foo": 2}\n' +
                               '{"time": "2016-01-01T00:02:00.000Z", "foo": 3}\n')

        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": 1},
            {"time": "2016-01-01T00:01:00.000Z", "foo": 2},
            {"time": "2016-01-01T00:02:00.000Z", "foo": 3}
        ])

    def test_stdio_can_read_a_single_timeless_point_(self):
        stdio.stdin = StringIO('{"foo": "bar"}')
        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"foo": "bar"}
        ])

    def test_stdio_can_read_multiple_timeless_points(self):
        stdio.stdin = StringIO('{"foo": 1}\n{"foo": 2}\n{"foo": 3}\n')
        results = []

        (
            read('stdio', format='jsonl')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"foo": 1},
            {"foo": 2},
            {"foo": 3}
        ])

    def test_stdio_can_read_multiple_points_with_custom_timefield(self):
        stdio.stdin = StringIO('{"bar": "2016-01-01T00:00:00.000Z", "foo": 1}\n' +
                               '{"bar": "2016-01-01T00:01:00.000Z", "foo": 2}\n' +
                               '{"bar": "2016-01-01T00:02:00.000Z", "foo": 3}\n')

        results = []

        (
            read('stdio', format='jsonl', time='bar')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {"time": "2016-01-01T00:00:00.000Z", "foo": 1},
            {"time": "2016-01-01T00:01:00.000Z", "foo": 2},
            {"time": "2016-01-01T00:02:00.000Z", "foo": 3}
        ])

    def test_stdio_can_write_nothing(self):
        stdio.stdin = StringIO('')
        stdio.stdout = StringIO('')
        (
            read('stdio')
            | write('stdio')
        ).execute()
        expect(stdio.stdout.getvalue()).to.eq('')

    def test_stdio_can_write_a_single_point(self):
        stdio.stdout = StringIO('')
        (
            emit(limit=1, start='2016-01-01')
            | write('stdio')
        ).execute()
        expect(json.loads(stdio.stdout.getvalue())).to.eq({
            'time': '2016-01-01T00:00:00.000Z'
        })

    def test_stdio_can_write_a_multiple_points(self):
        stdout = StringIO('')
        stdio.stdout = stdout
        (
            emit(limit=5, start='2016-01-01')
            | put(count=count())
            | write('stdio')
        ).execute()

        results = []
        stdio.stdin = StringIO(stdout.getvalue())
        (
            read('stdio')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'time': '2016-01-01T00:00:00.000Z', 'count': 1},
            {'time': '2016-01-01T00:00:01.000Z', 'count': 2},
            {'time': '2016-01-01T00:00:02.000Z', 'count': 3},
            {'time': '2016-01-01T00:00:03.000Z', 'count': 4},
            {'time': '2016-01-01T00:00:04.000Z', 'count': 5}
        ])

    def test_stdio_can_write_a_single_timeless_point(self):
        stdio.stdout = StringIO('')
        (
            emit(limit=1, start='2016-01-01')
            | put(foo='bar')
            | remove('time')
            | write('stdio')
        ).execute()
        expect(json.loads(stdio.stdout.getvalue())).to.eq({
            'foo': 'bar'
        })

    def test_stdio_can_write_a_multiple_timeless_points(self):
        stdout = StringIO('')
        stdio.stdout = stdout
        (
            emit(limit=5, start='2016-01-01')
            | put(count=count())
            | remove('time')
            | write('stdio')
        ).execute()

        results = []
        stdio.stdin = StringIO(stdout.getvalue())
        (
            read('stdio')
            | memory(results)
        ).execute()
        expect(results).to.eq([
            {'count': 1},
            {'count': 2},
            {'count': 3},
            {'count': 4},
            {'count': 5}
        ])



"""
read source
"""

from flume import adapters, node


class read(node):
    """
    read

    usage:
        read('adapter', [adapter arguments]) | ...

    The read source can be used to read points from various adapters of which
    include the following:
    """
    # XXX: dynamically attach to the docstrings the docstrings from various
    #      registered adapters ? 

    name = 'read'

    def __init__(self,
                 adapter,
                 **kwargs):
        node.__init__(self)
        cls = adapters.get_adapter(adapter, 'read')

        self.instance = cls(**kwargs)
        self.read = self.instance.read

    def loop(self):
        for points in self.read():
            self.push(points)
